import builtins
import copy
import dis
import imp
import io
import types
import marshal


class Frame:
    """Loosely based on CPython's Frame Object

    VERY incomplete.
    """
    UNDEF = object()

    def __init__(self, code, prev_frame=None,
                 globals=None, locals=None,
                 builtins=builtins.__dict__):
        self.f_back = prev_frame
        self.f_code = code
        self.f_lasti = None
        self.f_globals = globals if globals else {}
        self.f_locals = locals if locals else copy.copy(self.f_globals)
        self.f_builtins = builtins
        self.stack = []

    def exec(self, *args):
        # XXX Doesn't handle kwonly args
        assert len(args) == self.f_code.co_argcount
        assert self.f_code.co_kwonlyargcount == 0
        for i, arg in enumerate(args):
            if args is self.UNDEF:
                continue
            self.f_locals[self.f_code.co_varnames[i]] = arg
        co_code = self.f_code.co_code
        self.f_lasti = 0
        while True:
            opcode = co_code[self.f_lasti]
            meth = getattr(self, dis.opname[opcode])
            if opcode >= dis.HAVE_ARGUMENT:
                arg = co_code[self.f_lasti+1] + (co_code[self.f_lasti+2] << 8)
                meth(arg)
                self.f_lasti += 3
            else:
                meth()
                self.f_lasti += 1
            if opcode == dis.opmap['RETURN_VALUE']:
                return self.stack.pop()

    def LOAD_FAST(self, var_num):
        varname = self.f_code.co_varnames[var_num]
        self.stack.append(self.f_locals[varname])

    def STORE_FAST(self, var_num):
        varname = self.f_code.co_varnames[var_num]
        self.f_locals[varname] = self.stack.pop()

    def LOAD_CONST(self, consti):
        self.stack.append(self.f_code.co_consts[consti])

    def LOAD_NAME(self, namei):
        name = self.f_code.co_names[namei]
        if name in self.f_locals:
            val = self.f_locals[name]
        elif name in self.f_globals:
            val = self.f_globals[name]
        elif name in self.f_builtins:
            val = self.f_builtins[name]
        else:
            raise NameError("name '{}' is not defined".format(name))
        self.stack.append(val)

    def STORE_NAME(self, namei):
        # XXX CPython checks whether f_locals exists first
        name = self.f_code.co_names[namei]
        self.f_locals[name] = self.stack.pop()

    def COMPARE_OP(self, opname):
        op = dis.cmp_op[opname]
        arg0 = self.stack.pop()
        arg1 = self.stack.pop()
        source = '__crude_hack_arg0 {} __crude_hack_arg1'.format(op)
        locals = copy.copy(self.f_locals)
        assert '__crude_hack_arg0' not in locals
        assert '__crude_hack_arg1' not in locals
        locals['__crude_hack_arg0'] = arg0
        locals['__crude_hack_arg1'] = arg1
        res = eval(source, self.f_globals, locals)
        self.stack.append(res)

    def POP_TOP(self):
        self.stack.pop()

    def JUMP_FORWARD(self, delta):
        self.f_lasti += delta

    def POP_JUMP_IF_FALSE(self, target):
        if self.stack.pop():
            self.f_lasti = target - 3

    def BINARY_MULTIPLY(self):
        right = self.stack.pop()
        left = self.stack.pop()
        res = NotImplemented
        if hasattr(left, '__mul__'):
            res = left.__mul__(right)
        if res is NotImplemented and hasattr(right, '__mul__'):
            res = right.__mul__(left)
        if res is NotImplemented:
            raise TypeError(
                "unsuppored oprand type(s) for *: '{}' and '{}'".format(
                    left.__class__.__name__, right.__class__.__name__))
        self.stack.append(res)

    def BINARY_ADD(self):
        right = self.stack.pop()
        left = self.stack.pop()
        res = NotImplemented
        if hasattr(left, '__add__'):
            res = left.__add__(right)
        if res is NotImplemented and hasattr(right, '__add__'):
            res = right.__add__(left)
        if res is NotImplemented:
            raise TypeError(
                "usuppored operand type(s) for +: '{}' and '{}'".format(
                    left.__class__.__name__, right.__class__.__name__))
        self.stack.append(res)

    def MAKE_FUNCTION(self, argc):
        # XXX Doesn't handle default argument values
        cobj = self.stack.pop()
        func = types.FunctionType(cobj, self.f_globals)
        self.stack.append(func)

    def CALL_FUNCTION(self, argc):
        # XXX Doesn't handle default values or kwonly args
        nkwargs = argc >> 8
        nargs = argc & 0xff
        assert nkwargs == 0
        args = [self.stack.pop() for i in range(nargs)]
        args.reverse()
        func = self.stack.pop()
        if hasattr(func, '__code__'):
            frame = Frame(func.__code__, prev_frame=self,
                          globals=func.__globals__)
            ret = frame.exec(*args)
        else:
            ret = func.__call__(*args) # assume builtin function
        self.stack.append(ret)

    def RETURN_VALUE(self):
        pass




if __name__ == '__main__':
    with open(imp.cache_from_source('example.py'), 'rb') as fp:
        assert fp.read(4) == imp.get_magic()
        fp.seek(4, io.SEEK_CUR)
        code_mod = marshal.load(fp)
    code_sum = code_mod.co_consts[1]
    f_sum = Frame(code_sum)
    print (f_sum.exec(15, 4))
    f_mod = Frame(code_mod)
    f_mod.exec()
